import io
import os
import sys
import pkg_resources
import inquirer

sys.stdout = io.TextIOWrapper(sys.stdout.buffer, encoding='utf-8')
sys.stderr = io.TextIOWrapper(sys.stderr.buffer, encoding='utf-8')

system_choices = [
    inquirer.List(
        'system',
        message="Select system to use",
        choices=['powershell.exe', 'cmd.exe'],
    )
]

system_shortnames = {
    'powershell.exe': 'PS',
    'cmd.exe': 'CMD',
}

system_end_msgs = {
    'powershell.exe': '##END_OF_OUTPUT##',
    'cmd.exe': 'REM END_OF_OUTPUT',
}

system_pwd = {
    'powershell.exe': 'pwd | findstr :',
    'cmd.exe': 'cd',
}

def get_l2t_path():
    
        l2t_path = os.environ.get("L2T_PATH")
        if not l2t_path:
            os.environ["L2T_PATH"] = pkg_resources.resource_filename("llama2terminal", '')
        return l2t_path


from typing import Optional, List, Dict, Any 
from abc import abstractmethod, ABC
from dataclasses import dataclass

@dataclass
class Modules:
    CONVERSATIONAL_MODEL: str = 'You are a conversational model, always answer as the name as AI'
    ONE_LINE: str = "Answer always concisely in one line with a maximum of 20 words."
    SHORT_ANSWER: str = "Provide a detailed short answer in the range of 20-30 words."
    ONLY_CODE: str = "Only provide code fragments. Do not include any explanatory text."
    ELABORATE: str = "Please elaborate in detail, giving a comprehensive answer."
    YES_NO: str = "Answer with a simple 'yes' or 'no'."
    MULTIPLE_CHOICE: str = "Choose the most appropriate option: [A, B, C, D]."
    INSTRUCTION: str = "Give step-by-step instructions."
    SUMMARY: str = "Summarize the following information concisely."
    EXPLAIN_LIKE_5: str = "Explain it as if you're talking to a 5-year-old."
    PRO_CON: str = "List the pros and cons."
    EXAMPLES: str = "Provide examples to illustrate your point."
    COMPARISON: str = "Compare and contrast the following concepts."
    TIMELINE: str = "Describe the events in chronological order."
    PREDICTION: str = "Predict what will happen next based on the given information."
    PERSONAL_OPINION: str = "Give your personal opinion on the following matter, while being neutral and objective."
    FACT_CHECK: str = "Check the validity of the following statement and provide sources if possible."
    DO_NOT_MODIFY: str = "Do not use [] or add any modifications in this prompt"
    USE_TEMPLATE: str = "Alwayse response using this template: (AI_NAME): Response."


class DynamicParam(ABC):
    
    def __init__(self, name):
        self.name = name

    @abstractmethod
    def __str__(self) -> str:
        pass

class PromptTemplate:
    
    def __init__(
        self,
        name: Optional[str] = "LLM",
        history: Optional[str] = None,
        hard_params: Optional[Dict[str, Any]] = None,
        dynamic_params: Optional[DynamicParam] = None,
        modules: Optional[List[str]] = None,
        query: Optional[str] = ""
    ) -> None:
        self.name = name
        self.history = history
        self.hard_params = hard_params
        self.dynamic_params = dynamic_params
        self.modules = modules if modules else []
        self.query = query

    def __str__(self) -> str:

        prompt = f"[INST] Your name is {self.name}[INST]\n"
        
        if self.modules:
            module_list = '\n- '.join(self.modules)
            prompt += f"[MODULES] You MUST ALWAYS follow these rules:\n- {module_list}\n[MODULES]\n"


        if self.dynamic_params:
            for param in self.dynamic_params:
               prompt += f"[{param.name}] {str(param)} [\{param.name}]\n"

        if self.hard_params:
           for param, value in self.hard_params.items():
               prompt += f"[{param}] {value} [\{param}]\n"
        if self.history:
            prompt += f"[HISTORY] {self.history} [HISTORY]\n"
        
        prompt += f"[QUERY] {self.query} [QUERY]\n"
        
        return prompt

# Uso
# prompt = PromptTemplate(
#     name = "Sarah",
#     history="Miaw",
#     hard_params = {
#         'CODE': "set x as miaw",
#         'CODE2': "set y as mew"
#     },
#     modules = [
#        MODULES['SHORT_ANSWER'],
#        MODULES['ONLY_CODE'],
#        MODULES['EXAMPLES']
#     ],
#     query="Explain how to code a cat"
# )
# print(prompt.to_raw())



import os
import yaml
from llama2terminal.wrapper.terminal import ShellWrapper
from llama2terminal.wrapper.config import get_l2t_path
from llama2terminal.config.settings import generate_config

if __name__ == "__main__":
    
    l2t_path = get_l2t_path()
    config_path = os.path.join(l2t_path, "base", "config.yaml")
    with open(config_path, 'r') as file:
        config_data = yaml.safe_load(file)
    
    if config_data['general']['welcome']:
        print("It looks like its your first time here.")
        print("Loading Welcome settings...")
        config_data['general']['welcome'] = False
        with open(config_path, 'w') as outfile:
            yaml.safe_dump(config_data, outfile)
        generate_config()

    app = ShellWrapper()
    app.cmdloop()


import os
import logging
import yaml
import torch

from llama2terminal.base.prompts import PromptTemplate, Modules
from llama2terminal.wrapper.config import get_l2t_path

from abc import abstractmethod, ABC
from transformers import AutoModelForCausalLM, AutoTokenizer
from haystack.agents.conversational import ConversationalAgent
from haystack.agents.memory import NoMemory
from haystack.nodes import PromptNode

class LlamaAgent(ABC):

    def __init__(self):
        self.l2t_path = get_l2t_path()
        config_path = os.path.join(self.l2t_path, "base", "config.yaml")
        with open(config_path, 'r') as stream:
            self.config = yaml.safe_load(stream)

        logging.basicConfig(
            format="%(levelname)s - %(name)s -  %(message)s",
            level=self.config['general']['logging_level']
        )

        self.memory = []
        self.prompt_node = self.build_model()
        self.init_model()

    @abstractmethod
    def init_model(self):
        pass

    def build_model(self) -> PromptNode:

        os.environ['PYTORCH_CUDA_ALLOC_CONF'] = self.config['general']['pytorch_cuda_config']
        hf_token = self.config['model'].get('token')
        if not hf_token:
            hf_token = os.environ.get('HUGGINGFACE_TOKEN')
            if not hf_token or hf_token == 'HUGGINGFACE_TOKEN':
                raise ValueError("HUGGINGFACE_TOKEN not found in environment variable.")

        torch.cuda.empty_cache()

        # Building Model
        model_id = self.config['model']['id']
        max_length = self.config['model']['token_length']

        model = AutoModelForCausalLM.from_pretrained(
            model_id, load_in_4bit=True, use_auth_token=hf_token
        )
        model.config.pretraining_tp = 1
        tokenizer = AutoTokenizer.from_pretrained(model_id, use_auth_token=hf_token)

        pn = PromptNode(
            model_id,
            max_length=max_length,
            model_kwargs={
                'model': model,
                'tokenizer': tokenizer,
                'task_name': 'text2text-generation',
                'device': None,
                'stream': True
            }
        )

        return pn

    @abstractmethod
    def get_prediction(self, query) -> str:
        torch.cuda.empty_cache()
        output = self.prompt_node.run(query)['answers'][0].answer.split('\n')[-1]
        return output


class LlamaConversationalAgent(LlamaAgent):

    def init_model(self):
       self.agent = ConversationalAgent(
            prompt_node=self.prompt_node,
            memory=NoMemory()
        )

    def get_prediction(self, query) -> str:

        torch.cuda.empty_cache()
        history = '\n'.join(self.memory)

        prompt = PromptTemplate(
            name=self.config['agent']['name'],
            history=history,
            modules= [
                Modules.SHORT_ANSWER,
                Modules.ONE_LINE,
                Modules.DO_NOT_MODIFY,
                Modules.USE_TEMPLATE
            ],
            query=query
        )

        output = self.agent.run(str(prompt))['answers'][0].answer.split('\n')[-1]
        self.memory.append("Human:" + query)
        self.memory.append(output)

        return output


import cmd2
import os
import inquirer
import yaml


import llama2terminal.wrapper.config as config
from llama2terminal.wrapper.colors import TerminalColors
from llama2terminal.wrapper.history import CommandLogger
from llama2terminal.wrapper.IOUtils import CommandLineReader
from llama2terminal.packages.load import package_loader

class ShellWrapper(cmd2.Cmd):

    def __init__(self):

        super().__init__()

        self.l2t_path = config.get_l2t_path()
        yaml_path = os.path.join(self.l2t_path, "wrapper", "params.yaml")
        with open(yaml_path) as file:
            self.params = yaml.safe_load(file)

        self.listening = False
        self.sys_type = self.params['DEFAULT']['system']
        self.cmd_logger = CommandLogger()
        self.clr = CommandLineReader(self.sys_type)
        self.change_prompt()

    def change_prompt(self):
        color = TerminalColors.GREEN if self.listening else TerminalColors.ENDC
        pwd = self.clr.get_pwd().replace('\n','')
        self.prompt = f"[{color}Llama2{TerminalColors.ENDC}] {config.system_shortnames[self.sys_type]} {pwd}> "
    
    def default(self, statement):
        output, error = self.clr.run_command(statement.raw)
        if len(output) > 0:
            self.poutput(output)
        if len(error) > 0:
            self.perror(error)
        if self.listening:
            self.cmd_logger.log_command(statement.raw, output=output, error=error)
        self.change_prompt()

    def do_exit(self, args):
        self.clr.close()
        return True

    def do_llama(self, args):

        split_args = args.split()
        if not split_args:
            return

        match split_args[0]:
            case "listen":
                self.listening = True
            case "pause":
                self.listening = False
            case "sys":
                self.sys_type = inquirer.prompt(config.system_choices)['system']
                self.clr.close()
                self.clr = CommandLineReader(self.sys_type)
            case "log":
                self.cmd_logger.display_log()
            case "clear":
                self.cmd_logger.clear()
            case "stop":
                self.clr.close()
                return True
            case _:
                try:
                    package_loader.run_module(split_args[0],split_args[1:])
                except ValueError as e:
                    self.perror(e)
        
        self.change_prompt()

                



import yaml
import os
import importlib.util

from llama2terminal.wrapper.config import get_l2t_path

class PackageLoader:

    def __init__(self):

        self.l2t_path = get_l2t_path()
        yaml_path = os.path.join(self.l2t_path, "packages", "pkg.yaml")

        with open(yaml_path, "r") as file:
            self.packages = yaml.safe_load(file)

        self.loaded_modules = {}

    def unload_module(self, package_name):
        if package_name in self.packages:
            del self.packages[package_name]

    def load_initial_modules(self):
        for package_name, package_info in self.packages.items():
            if package_info.get("load_at_start", False):
                self.run_module(package_name)

    def load_module(self, package_name):

        if package_name in self.loaded_modules:
            return
        
        # Verify package is installled 
        if package_name not in self.packages:
            raise ValueError(f"Package '{package_name}' is not installed.")
        
        package_dir = self.packages[package_name]["dir"]

        # Load "launch.py"
        spec = importlib.util.spec_from_file_location("launch_module", os.path.join(self.l2t_path, "packages", package_dir, "launch.py"))
        launch_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(launch_module)
        
        # Save the package to loaded modules
        self.loaded_modules[package_name] = launch_module

    def run_module(self, package_name, *args):
        """
        Executes the previously loaded package.

        Args:
        - package_name (str): The name of the package you wish to run, e.g., "talk".
        - *args, **kwargs: Arguments and keyword arguments for the 'start' function of the module.
        """

        if package_name not in self.loaded_modules:
            self.load_module(package_name)
        
        package_dir = self.packages.get(package_name, {}).get("dir")
        if not package_dir:
            raise ValueError(f"Directory for package {package_name} has not been found.")

        # Launch run.py
        spec = importlib.util.spec_from_file_location("run_module", os.path.join(self.l2t_path, "packages", package_dir, "run.py"))
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        module.__start__(*args)

package_loader = PackageLoader()


from datetime import datetime
from typing import List, Optional
from llama2terminal.wrapper.colors import TerminalColors

class CommandEntry:
    def __init__(self, command: str):
        self.command = command
        self.timestamp = datetime.now()
        self.output = None
        self.error = None

    def set_output(self, output: str):
        self.output = output

    def set_error(self, error: str):
        self.error = error

    def to_raw(self):
        return f"<[TIMESTAMP]{self.timestamp}[\TIMESTAMP]\n" + \
                f"[COMMAND]{self.command}[\COMMAND]\n" + \
                f"[OUTPUT]{self.output}[\OUTPUT]\n" + \
                f"[ERROR]{self.error}[\ERROR]>"

    def __str__(self):
        return f"{TerminalColors.ORANGE}{self.timestamp}: {self.command}\n \
            {TerminalColors.GREEN}Output:\n{TerminalColors.ENDC}{self.output}\n \
            {TerminalColors.WARNING}Error:\n{self.error} {TerminalColors.ENDC}"

class CommandLogger:
    def __init__(self):
        self.commands: List[CommandEntry] = []

    def log_command(self, command: str, output: Optional[str] = None, error: Optional[str] = None):
        entry = CommandEntry(command)
        if output:
            entry.set_output(output)
        if error:
            entry.set_error(error)
        self.commands.append(entry)

    def clear(self):
        self.commands.clear()

    def to_raw(self) -> List[str]:
        raw_logs = []
        for entry in self.commands:
            raw_logs.append(entry.to_raw())
        return raw_logs

    def display_log(self):
        for entry in self.commands:
            print(entry)
            print('=' * 50)


from setuptools import setup, find_packages

setup(
    name="llama2-terminal",
    version="0.1.1",
    packages=find_packages(),
    install_requires=[
        "inquirer",
        "cmd2",
        "farm-haystack",
        "torch",
        "transformers>=4.0.0",
        "bitsandbytes>=0.3.9",
        "pyyaml"
    ],
    author="SamthinkGit",
    author_email="sebastianmayorquin@gmail.com",
    description="Descripción breve del proyecto",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/SamthinkGit/llama2-terminal",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
)



from llama2terminal.base.agents import LlamaConversationalAgent

agent = LlamaConversationalAgent()
while True:
    query = input("Query: ")
    if query == "exit":
        break
    print(agent.get_prediction(query))
    


import subprocess
import os
import threading
import llama2terminal.wrapper.config as config
from queue import Queue

class CommandLineReader:
    
    def __init__(self, sys_type):
        self.sys_type = sys_type
        self.stderr_queue = Queue()
        self._end_output_signal = config.system_end_msgs[sys_type]
        self.start()

    def start(self):
        self.shell_process = subprocess.Popen([self.sys_type], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        self.stderr_thread = threading.Thread(target=self.error_reader)
        self.stop_event = threading.Event()
        self.stderr_thread.start()

        # Run a command for fully starting the process
        self.run_command("")

    def error_reader(self):
        while not self.stop_event.is_set():
            line = self.shell_process.stderr.readline()
            if line:
                self.stderr_queue.put(line)

    def get_pwd(self):
        output, _ = self.run_command(config.system_pwd[self.sys_type])
        return output
        
    def run_command(self, cmd):

        # Send the command to the process
        delimiter = self._end_output_signal
        self.shell_process.stdin.write(cmd + '\n')
        self.shell_process.stdin.write(delimiter + '\n')
        self.shell_process.stdin.flush()

        # Remove the query line\s
        line = ""
        while cmd not in line:
            line = self.shell_process.stdout.readline()

        # Read stdout until delimeter 
        output_lines = []
        while True:
            line = self.shell_process.stdout.readline()
            if not line:
                continue
            if delimiter in line:
                break
            if line:
                output_lines.append(line) 

        # Read non-blocking errors
        error_lines = []
        while not self.stderr_queue.empty():
            line = self.stderr_queue.get_nowait()
            if line:
                error_lines.append(line)

        # Flush other outputs 
        self.shell_process.stdout.flush()
        self.shell_process.stderr.flush()

        return ''.join(output_lines), ''.join(error_lines)

    def close(self):
        if self.shell_process:
            self.stop_event.set()
            self.shell_process.terminate()
            self.stderr_thread.join()
